import asyncio
import inspect
import logging
from typing import Optional

from aiogram.types import Message, CallbackQuery, Update

from aiogram_metrics.api import handle_event
from aiogram_metrics.hub import Hub


def track(arg=None):
    """
    Decorator for saving events. Can be used either with or without an argument.
    If an argument is not provided, message text or callback data will be extracted.
    :param arg: Event name for tracking
    """
    _func = None

    if arg and callable(arg):
        _func = arg
        event = None
    else:
        event = arg

    def decorator(func):

        async def wrapper(*args, **kwargs):
            update = kwargs.get('event_update')

            # Extract argument names
            sig = inspect.signature(func)
            kwargs = {k: v for k, v in kwargs.items() if k in sig.parameters}

            if not update:
                logging.info(f"No update found for event '{event}', skipping...")
            else:
                if Hub.is_activated:
                    try:
                        coro = handle_event(event, update=update)
                        asyncio.create_task(coro)
                    except Exception as e:
                        logging.error(f"Failed to handle event: {e}")
                else:
                    logging.warning('aiogram-metrics is not registered!')

            if not asyncio.iscoroutinefunction(func):
                async def async_wrapper():
                    return await func(*args, **kwargs)
                return await async_wrapper()
            else:
                return await func(*args, **kwargs)

        return wrapper

    return decorator(_func) if _func else decorator


def manual_track(
    event: Optional[str] = None,
    *,
    update: Optional[Update] = None,
    message: Optional[Message] = None,
    callback_query: Optional[CallbackQuery] = None,
    user_id: Optional[int] = None,
):
    """
    Manual event tracking. Sane as `track` decorator
    """
    coro = handle_event(event, update=update, message=message, callback_query=callback_query, user_id=user_id)
    asyncio.create_task(coro)



import json
import logging
from datetime import datetime as dt
from typing import Optional

from aiogram import Dispatcher
from aiogram.types import Update, User, Message, CallbackQuery

try:
    from aiogram.contrib.middlewares.i18n import I18nMiddleware
except ModuleNotFoundError:
    from aiogram.utils.i18n.middleware import I18nMiddleware

from psycopg import AsyncConnection
from psycopg.sql import Identifier

from aiogram_metrics.sql import MessageType, init_table, save_event
from aiogram_metrics.hub import Hub


__all__ = [
    'register',
    'close',
    'handle_event'
]


async def register(dsn: str, table_name: str):
    Hub.logger = logging.getLogger('aiogram-metrics')

    Hub.logger.debug('Checking database connection...')
    Hub.connection = await AsyncConnection.connect(dsn)

    Hub.logger.debug(f'Checking table {table_name}...')
    Hub.table_name = Identifier(table_name).as_string(Hub.connection)
    await init_table()
    Hub.logger.info('Registration complete! Waiting for events...')

    Hub.is_activated = True


async def close():
    if Hub.connection:
        await Hub.connection.close()
        Hub.logger.info('Connection is now closed. Bue!')


async def handle_event(
    event: Optional[str] = None,
    *,
    update: Optional[Update] = None,
    message: Optional[Message] = None,
    callback_query: Optional[CallbackQuery] = None,
    user_id: Optional[int] = None,
):
    if not Hub.is_activated:
        logging.warning('aiogram-metrics is not registered!')
        return

    if not update and not message and not callback_query:
        Hub.logger.info('No update found! Skipping...')

    if (update and update.message) or message:
        message = message or update.message
        user_id = message.chat.id
        message_id = message.message_id
        message_data = {'text': message.text}
        message_type = (message.text.startswith('/') and MessageType.command) or MessageType.message
        event = event or message.text
        language = message.from_user.language_code
    elif (update and update.callback_query) or callback_query:
        callback_query = callback_query or update.callback_query
        user_id = callback_query.message.chat.id
        message_id = callback_query.message.message_id
        message_data = {'callback_data': callback_query.data}
        message_type = MessageType.callback
        event = event or callback_query.data or callback_query.message.text
        language = callback_query.from_user.language_code
    elif user_id:
        message_id = language = None
        message_data = {}
        message_type = MessageType.none
    else:
        raise ValueError('Unknown message type! Currently supported `message` and `callback_query`.')

    message_data = json.dumps(message_data, ensure_ascii=False, indent=2)
    event_data = (event, dt.now().isoformat(), user_id, message_id, message_type, message_data, language)
    await save_event(event_data)

    Hub.logger.debug(f'Successfully saved event "{event}"!')
